// Copyright 2002-2003 University of Guelph, All Rights Reserved
import java.io.*;
import java.util.*;

class NewKProb {

  public PrintWriter out;
  public PrintWriter solnOut;
  public int type;
  public int solution[];
  public Random solutionRand;
  public ArrayList allowedRelations[];
  public Random varRand;
  public Random clueRand;
  public Problem problem;
  public HashSet knownVars;
  public int given[];

  NewKProb(PrintWriter Out, PrintWriter SolnOut, int Type) {
	out = Out;
	solnOut = SolnOut;
	type = Type;
  }

  public void generateProblem() throws Exception {
	problem = new Problem(type);
	Problem solnProblem = new Problem(type);

	createUniqueConstraints();
	solutionRand = new Random();
	clueRand = new Random();
	varRand = new Random();
	generateSolution();
	copySoln(solnProblem);
	solnOut.print(solnProblem.getSolutionString());
	solnProblem = null;

	allowedRelations = new ArrayList[2];
	allowedRelations[0] = new ArrayList();
	allowedRelations[1] = new ArrayList();

	if (type == Problem.ZEBRA) {
	  if (Debug.DEBUG_LEVEL >= Debug.DEBUG_MEDIUM) {
		System.out.println("Only allowing Zebra-type relations");
	  }
	  allowedRelations[0].add(new Integer(Problem.SAME_COL));
	  allowedRelations[1].add(new Double(0.66));
	  allowedRelations[0].add(new Integer(Problem.NEXT_TO));
	  allowedRelations[1].add(new Double(0.25));
	  allowedRelations[0].add(new Integer(Problem.NEXT_RIGHT));
	  allowedRelations[1].add(new Double(0.08));
	  allowedRelations[0].add(new Integer(Problem.GIVEN));
	  allowedRelations[1].add(new Double(0.01));
	} else {
	  if (Debug.DEBUG_LEVEL >= Debug.DEBUG_MEDIUM) {
		System.out.println("Using Sherlock type relations");
	  }
	  allowedRelations[0].add(new Integer(Problem.SAME_COL));
	  allowedRelations[1].add(new Double(0.35));
	  allowedRelations[0].add(new Integer(Problem.NOT_SAME_COL));
	  allowedRelations[1].add(new Double(0.05));
	  allowedRelations[0].add(new Integer(Problem.NEXT_TO));
	  allowedRelations[1].add(new Double(0.35));
	  allowedRelations[0].add(new Integer(Problem.NOT_NEXT_TO));
	  allowedRelations[1].add(new Double(0.04));
	  allowedRelations[0].add(new Integer(Problem.LEFT_OF));
	  allowedRelations[1].add(new Double(0.04));
	  allowedRelations[0].add(new Integer(Problem.RIGHT_OF));
	  allowedRelations[1].add(new Double(0.04));
	  allowedRelations[0].add(new Integer(Problem.NOT_NEXT_TO_SAME));
	  allowedRelations[1].add(new Double(0.01));
	  allowedRelations[0].add(new Integer(Problem.NEXT_RIGHT));
	  allowedRelations[1].add(new Double(0.04));
	  allowedRelations[0].add(new Integer(Problem.NEXT_LEFT));
	  allowedRelations[1].add(new Double(0.04));
	  allowedRelations[0].add(new Integer(Problem.GIVEN));
	  allowedRelations[1].add(new Double(0.09));
	}

	generateClues();
	out.print(relationsString());
  }

  public void copySoln(Problem problem) {
	for (int i = 0; i <= problem.getMaxVars(); i++) {
	  problem.solution[i] = solution[i];
	}
  }

  public boolean hasOneSolution() {
	problem.arc3();
	boolean done = true;
	for (int i = 1; i < problem.domain.length; i++) {
	  if (((HashSet)(problem.domain[i])).size() > 1) {
		done = false;
	  } else {
		knownVars.add(new Integer(i));
	  }
	}
	return done;
  }  

  public void generateClues() throws Exception {
	Random randVar = new Random();
	//	HashSet relations = new HashSet();
	knownVars = new HashSet();
	given = new int[problem.getMaxVars() + 1];
	for (int i = 0; i <= problem.getMaxVars(); i++) {
	  given[i] = 0;
	}

	// Start by choosing two variables as given (domain = solution only)
	int x1 = randVar.nextInt(problem.getMaxVars()) + 1;
	int x2 = randVar.nextInt(problem.getMaxVars()) + 1;

	problem.domain[x1] = new HashSet();
	knownVars.add(new Integer(x1));
	problem.domain[x1].add(new Integer(solution[x1]));
	while (x2 == x1) {
	  x2 = randVar.nextInt(problem.getMaxVars()) + 1;
	}	
	problem.domain[x2] = new HashSet();
	knownVars.add(new Integer(x2));
	problem.domain[x2].add(new Integer(solution[x2]));
	given[x1] = solution[x1];
	given[x2] = solution[x2];

	while (!(hasOneSolution())) {
	  getClue();
	}
  }

  public  boolean checkAllowed(int i, int j, int clueType) {
	int Vi, Vj;
	boolean result = false;

	if (problem.constraints[i][j] != null) {
	  result = false;
	} else {

	  Vi = solution[i];
	  Vj = solution[j];
	  
	  switch (clueType) {
		case Problem.SAME_COL:		
		  if (Vi == Vj) {
			result = true;
		  } 
		  break;
		case Problem.NOT_SAME_COL:
		  if (Vi != Vj) {
			result = true;
		  }
		  break;
		case Problem.NEXT_TO:
		  if (((Vi - Vj) ==  1) || 
			  ((Vi - Vj) == -1)) {
			result = true;
		  }
		  break;	
		case Problem.NOT_NEXT_TO:
		  if (((Vi - Vj) !=  1) &&
			  ((Vi - Vj) != -1)) {
			result = true;
		  }
		  break;
		case Problem.LEFT_OF:
		  if (Vi < Vj) {
			result = true;
		  }
		  break;
		case Problem.RIGHT_OF:
		  if (Vi > Vj) {
			result = true;
		  }
		  break;
		case Problem.NOT_EQUAL:
		  if (Vi != Vj) {
			result = true;
		  }
		  break;
		case Problem.NOT_NEXT_TO_SAME:
		  if (((Vi - Vj) !=  1) &&
			  ((Vi - Vj) != -1) &&
			  ((Vi != Vj)))
			{
			  result = true;
			}
		  break;
		case Problem.NEXT_RIGHT:
		  if ((Vi - Vj) ==  1) {		  
			result = true;
		  }
		  break;
		case Problem.NEXT_LEFT:
		  if ((Vi - Vj) ==  -1) { 
			result = true;
		  }
		  break;
		case Problem.GIVEN:
		  result = true;
		  break;
	  }  
	}
	return result;
  }

  public  HashSet getRowSet(int v) {
	HashSet newSet = new HashSet();
	int numRows = (int)(Math.sqrt(problem.getMaxVars()));
	int rowNum = (v - 1) / numRows;

	for (int i = 0; i < numRows; i++) {
	  newSet.add(new Integer(rowNum * numRows + i + 1));
	}
	return newSet;
  }

  public  int getClueType() {
	double clueDouble = clueRand.nextDouble();
	double lowAllowedDouble = 0;
	double highAllowedDouble = ((Double)(allowedRelations[1].get(0))).doubleValue();
	for (int i = 0; i < allowedRelations[1].size(); i++) {
	  if (Debug.DEBUG_LEVEL >= Debug.DEBUG_VERBOSE) {
		System.out.println("i: " + String.valueOf(i) + " low: " +
						   String.valueOf(lowAllowedDouble) + " high: " +
						   String.valueOf(highAllowedDouble) + " rand: " +
						   String.valueOf(clueDouble));
	  }
	  if ((clueDouble >= lowAllowedDouble) && (clueDouble <= highAllowedDouble)) {
		int clueType = ((Integer)(allowedRelations[0].get(i))).intValue();
		return clueType;
	  }
	  if (i == (allowedRelations[1].size() - 1)) {
		return -2;
	  }
		lowAllowedDouble += ((Double)(allowedRelations[1].get(i))).doubleValue();
		highAllowedDouble += ((Double)(allowedRelations[1].get(i + 1))).doubleValue();
	}
	return -1;
  }

  public  Object getFromSet(Set set, int index) {
	int i = 0;
	Iterator itemIterator = set.iterator();
	Object item = null;

	while (itemIterator.hasNext() && i <= index) {
	  item = itemIterator.next();
	  i++;
	}
	if (Debug.DEBUG_LEVEL >= Debug.DEBUG_VERBOSE) {
	  System.out.println("Item " + item + " i: " + String.valueOf(i) + 
						 " index: " + String.valueOf(index));
	}

	return item;
  }

  public  ArrayList getUnknownVars() {
	ArrayList unknownVars = new ArrayList();

	for (int i = 1; i <= problem.getMaxVars(); i++) {
	  unknownVars.add(new Integer(i));
	}
	unknownVars.removeAll(knownVars);

	return unknownVars;

  }

  public void getClue() throws Exception {
	int clueType = getClueType();
	ArrayList unknownVars = getUnknownVars();
	Iterator kvIterator = knownVars.iterator();
	Integer vxInteger;
	int vx;
	int vy;
	HashSet possibleClues = new HashSet();

	// Figure out what clues are possible using one known and one unknown 
	// variable for the given clue type
	while (kvIterator.hasNext()) {
	  vxInteger = (Integer)(kvIterator.next());
	  vx = vxInteger.intValue();

	  for (int i = 0; i < unknownVars.size(); i++) {
		vy = ((Integer)(unknownVars.get(i))).intValue();
		if (checkAllowed(vx, vy, clueType)) {
		  possibleClues.add(new ArcObj(vx, vy));
		}
	  }
	}

  	// If no matches try unknown only
	if (possibleClues.size() == 0) {
	  for (int i = 0; i < unknownVars.size(); i++) {
		for (int j = 0; j < unknownVars.size(); j++) {
		  vx = ((Integer)(unknownVars.get(i))).intValue();
		  vy = ((Integer)(unknownVars.get(j))).intValue();
		  if (vx == vy) {
			continue;
		  } else {
			if (checkAllowed(vx, vy, clueType)) {
			  possibleClues.add(new ArcObj(vx, vy));
			}
		  }
		}
	  }
	}	

   	// If still no matches try known only
	if (possibleClues.size() == 0) {
	  Iterator kvYIterator = knownVars.iterator();
	  Integer vyInteger;
	  
	  while (kvIterator.hasNext()) {
		vxInteger = (Integer)(kvIterator.next());
		vx = vxInteger.intValue();

		while (kvYIterator.hasNext()) {
		  vyInteger = (Integer)(kvYIterator.next());
		  vy = vyInteger.intValue();
		  if (vx == vy) {
			continue;
		  } else {
			if (checkAllowed(vx, vy, clueType)) {
			  possibleClues.add(new ArcObj(vx, vy));
			}
		  }
		}
	  }
	}	
	

	// If still no matches throw an exception
	if (possibleClues.size() == 0) {
	  throw (new Exception("No possible matches for clue: something is badly wrong." + String.valueOf(clueType)));
	}

	int randChoice = varRand.nextInt(possibleClues.size());

	if (Debug.DEBUG_LEVEL >= Debug.DEBUG_HIGH) {
	  System.out.println("Choosing clue " + String.valueOf(randChoice) + 
						 " of " + String.valueOf(possibleClues.size()) + 
						 " clues.");

	}
	ArcObj newClue = (ArcObj)(getFromSet(possibleClues, randChoice));

	if (clueType == Problem.GIVEN) {
	  given[newClue.j] = solution[newClue.j];
	  problem.domain[newClue.j] = new HashSet();
	  problem.domain[newClue.j].add(new Integer(solution[newClue.j]));
	} else {
	  problem.addConstraint(newClue.i, newClue.j, clueType);
	}

	switch (clueType) {
	  case Problem.SAME_COL:		
		problem.addConstraint(newClue.j, newClue.i, clueType);
		break;
	  case Problem.NOT_SAME_COL:
		problem.addConstraint(newClue.j, newClue.i, clueType);
		break;
	  case Problem.NEXT_TO:
		problem.addConstraint(newClue.j, newClue.i, clueType);
		break;	
	  case Problem.NOT_NEXT_TO:
		problem.addConstraint(newClue.j, newClue.i, clueType);
		break;
	  case Problem.LEFT_OF:
		problem.addConstraint(newClue.j, newClue.i, Problem.RIGHT_OF);
		break;
	  case Problem.RIGHT_OF:
		problem.addConstraint(newClue.j, newClue.i, Problem.LEFT_OF);
		break;
	  case Problem.NOT_EQUAL:
		problem.addConstraint(newClue.j, newClue.i, clueType);
		break;
	  case Problem.NOT_NEXT_TO_SAME:
		problem.addConstraint(newClue.j, newClue.i, clueType);
		break;
	  case Problem.NEXT_RIGHT:
		problem.addConstraint(newClue.j, newClue.i, Problem.NEXT_LEFT);
		break;
	  case Problem.NEXT_LEFT:
		problem.addConstraint(newClue.j, newClue.i, Problem.NEXT_RIGHT);
		break;
	}
  }

  public  Vector doPermute(TreeSet curDomain) {
	Vector permutations = new Vector();
	Vector integers;

	if (curDomain.size() == 1) {
	  integers = new Vector();
	  integers.addElement((Integer)(curDomain.first()));
	  permutations.addElement(integers);
	} else {	  
	  Iterator current = curDomain.iterator();	 
	  while (current.hasNext()) {
		TreeSet newDomain = (TreeSet)(curDomain.clone());
		Integer curInt = (Integer)(current.next());
		newDomain.remove(curInt);

		Vector endCombos = doPermute(newDomain);
		for (int i = 0; i < endCombos.size(); i++) {
		  integers = new Vector();
		  integers.addElement(curInt);
		  Vector endInts = (Vector)(endCombos.elementAt(i));

		  for (int j = 0; j < endInts.size(); j++) {
			integers.addElement((Integer)(endInts.elementAt(j)));
		  }
		  permutations.addElement(integers);
		}		
	  }	  
	}

	if (Debug.DEBUG_LEVEL >= Debug.DEBUG_MAX) {
	  for (int i = 0; i < permutations.size(); i++) {
		Vector ints = (Vector)(permutations.elementAt(i));
		System.out.print(String.valueOf(i) + ": ");
		for (int j = 0; j < ints.size(); j++) {
		  System.out.print(String.valueOf((Integer)(ints.elementAt(j)) + " "));
		}
		System.out.println("");
	  }
	}
	return permutations;
  }

  public  void setSolution(Object permVector, int varStart) {
	Vector permutes = (Vector)(permVector);
	for (int o = 0; o < permutes.size(); o++) {
	  solution[varStart + o] = 
		((Integer)(permutes.elementAt(o))).intValue();
	}
  }

  public  void generateSolution() {
	solution = new int[problem.getMaxVars() + 1];

	for (int i = 0; i <= problem.getMaxVars(); i++) {
	  solution[i] = Problem.NO_SOLUTION;
	}
	
	TreeSet topDomain = new TreeSet();
	Vector varPermutations;

	for (int i = 1; i <= problem.getMaxDomain(); i++) {
	  topDomain.add(new Integer(i));
	}

	varPermutations = doPermute(topDomain);

	for (int i = 0; i < problem.getMaxDomain(); i++) {
	  int x;
	  x = solutionRand.nextInt(varPermutations.size());
	  
	  setSolution(varPermutations.elementAt(x), i * problem.getMaxDomain() + 1 );
	}	
  }

  public void createUniqueConstraints() {
	HashSet rowSet;
	int v2;

	for (int i = 1; i <= problem.getMaxVars(); i++) {
	  rowSet = getRowSet(i);	  

	  Iterator rowIterator = rowSet.iterator();
	  while (rowIterator.hasNext()) {
		v2 = ((Integer)(rowIterator.next())).intValue();
		if (i != v2) {
		  problem.addConstraint(i, v2, Problem.NOT_EQUAL);
		}
	  }
	}
  }

  public String relationsString() {
	String out = "";

	if (type == Problem.SHERLOCK) {
	  out += "S\n";
	} else if (type == Problem.ZEBRA) {
	  out += "Z\n";
	}

	for (int i = 1; i <= problem.getMaxVars(); i++) {
	  if (given[i] != 0) {
		out += String.valueOf(i) + " is " + String.valueOf(given[i]) + "\n";
	  }
	}

	for (int i = 1; i <= problem.getMaxVars(); i++) {
	  for (int j = 1; j <= problem.getMaxVars(); j++) {
		if ((i != j) && (problem.constraints[i][j] != null)) {
		  out += problem.getConstraintString(i, j);
		}
	  }
	}
	return out;
  }
}


