// Copyright 2002-2003 University of Guelph, All Rights Reserved
import java.io.*;
import java.util.*;

class NewNonKProb {

  public PrintWriter out;
  public PrintWriter solnOut;
  public int type;
  public int solution[];
  public Random solutionRand;
  public ArrayList allowedRelations[];
  public Random varRand;
  public Random clueRand;
  public Problem problem;
  public int given[];
  public int colExtraUsed;
  public int rowExtraUsed;

  public HashMap knownRows;
  public HashMap knownCols;
  //  public boolean done;
  public HashSet knownVars;
  public int extraVarUse;
  

  NewNonKProb(PrintWriter Out, PrintWriter SolnOut, int Type) {
	out = Out;
	solnOut = SolnOut;
	type = Type;
  }

  public void generateProblem() throws Exception {
	problem = new Problem(type);
	//	done = false;
	Problem solnProblem = new Problem(type);

	createUniqueConstraints();
	solutionRand = new Random();
	clueRand = new Random();
	varRand = new Random();
	generateSolution();
	copySoln(solnProblem);
	solnOut.print(solnProblem.getSolutionString());
	solnProblem = null;

	allowedRelations = new ArrayList[2];
	allowedRelations[0] = new ArrayList();
	allowedRelations[1] = new ArrayList();

	if (type == Problem.ZEBRA) {
	  if (Debug.DEBUG_LEVEL >= Debug.DEBUG_MEDIUM) {
		System.out.println("Only allowing Zebra-type relations");
	  }
	  allowedRelations[0].add(new Integer(Problem.SAME_COL));
	  allowedRelations[1].add(new Double(0.66));
	  allowedRelations[0].add(new Integer(Problem.NEXT_TO));
	  allowedRelations[1].add(new Double(0.25));
	  allowedRelations[0].add(new Integer(Problem.NEXT_RIGHT));
	  allowedRelations[1].add(new Double(0.08));
	  allowedRelations[0].add(new Integer(Problem.GIVEN));
	  allowedRelations[1].add(new Double(0.01));
	} else {
	  if (Debug.DEBUG_LEVEL >= Debug.DEBUG_MEDIUM) {
		System.out.println("Using Sherlock type relations");
	  }
	  allowedRelations[0].add(new Integer(Problem.SAME_COL));
	  allowedRelations[1].add(new Double(0.35));
	  allowedRelations[0].add(new Integer(Problem.NOT_SAME_COL));
	  allowedRelations[1].add(new Double(0.05));
	  allowedRelations[0].add(new Integer(Problem.NEXT_TO));
	  allowedRelations[1].add(new Double(0.35));
	  allowedRelations[0].add(new Integer(Problem.NOT_NEXT_TO));
	  allowedRelations[1].add(new Double(0.04));
	  allowedRelations[0].add(new Integer(Problem.LEFT_OF));
	  allowedRelations[1].add(new Double(0.04));
	  allowedRelations[0].add(new Integer(Problem.RIGHT_OF));
	  allowedRelations[1].add(new Double(0.04));
	  allowedRelations[0].add(new Integer(Problem.NOT_NEXT_TO_SAME));
	  allowedRelations[1].add(new Double(0.01));
	  allowedRelations[0].add(new Integer(Problem.NEXT_RIGHT));
	  allowedRelations[1].add(new Double(0.04));
	  allowedRelations[0].add(new Integer(Problem.NEXT_LEFT));
	  allowedRelations[1].add(new Double(0.04));
	  allowedRelations[0].add(new Integer(Problem.GIVEN));
	  allowedRelations[1].add(new Double(0.09));
	}
	for (int i = 0; i <= problem.getMaxVars(); i++) {
	  problem.domain[i] = new HashSet();
	  for (int j = 1; j <= problem.getMaxDomain(); j++) {
		problem.domain[i].add(new Integer(j));
	  }
	}

	generateClues();
	out.print(relationsString());
  }

  public void copySoln(Problem problem) {
	for (int i = 0; i <= problem.getMaxVars(); i++) {
	  problem.solution[i] = solution[i];
	}
  }

  public boolean hasOneSolution() {
// 	problem.arc3();
// 	boolean done = true;
// 	for (int i = 1; i < problem.domain.length; i++) {
// 	  if (((HashSet)(problem.domain[i])).size() > 1) {
// 		done = false;
// 	  } else {
// 		knownVars.add(new Integer(i));
// 	  }
// 	}
// 	return done;
//	return done;
	HashSet domain[] = new HashSet[problem.getMaxVars() + 1];
	for (int i = 0; i <= problem.getMaxVars(); i++) {
	  domain[i] = (HashSet)(problem.domain[i].clone());
	}
	int result = bcssp(problem.getMaxVars());
	if (result == SOLUTION) {
	  return true;
	} else if (result == NO_SOLUTION) {
	  System.err.println("No solution possible.");
	} else if (result == MORE_THAN_ONE_SOLUTION) {
	} 
	for (int i = 0; i <= problem.getMaxVars(); i++) {
	  problem.domain[i] = (HashSet)(domain[i].clone());
	}
	
	return false;
  }  

  public void generateClues() throws Exception {
	Random randVar = new Random();
	//	HashSet relations = new HashSet();
	knownRows = new HashMap();
	knownCols = new HashMap();
	knownVars = new HashSet();
	for (int i = 0; i <= problem.getMaxDomain(); i++) {
	  knownRows.put(new Integer(i), new Integer(0));
	  knownCols.put(new Integer(i), new Integer(0));
	}

	given = new int[problem.getMaxVars() + 1];
	for (int i = 0; i <= problem.getMaxVars(); i++) {
	  given[i] = 0;
	}

	// Start by choosing two variables as given (domain = solution only)
	int x1 = randVar.nextInt(problem.getMaxVars()) + 1;
	int x2 = randVar.nextInt(problem.getMaxVars()) + 1;

	problem.domain[x1] = new HashSet();
	knownRows.put(new Integer(problem.varRow(x1)), new Integer(1));
	problem.domain[x1].add(new Integer(solution[x1]));
	while (x2 == x1) {
	  x2 = randVar.nextInt(problem.getMaxVars()) + 1;
	}	
	problem.domain[x2] = new HashSet();
	knownRows.put(new Integer(problem.varRow(x2)), new Integer(1));
	problem.domain[x2].add(new Integer(solution[x2]));
	given[x1] = solution[x1];
	given[x2] = solution[x2];

	colExtraUsed = 0;
	rowExtraUsed = 0;
	extraVarUse = 0;

	while (!(hasOneSolution())) {
	  getClue();
	}
  }

  public  boolean checkAllowed(int i, int j, int clueType) {
	int Vi, Vj;
	boolean result = false;

	Vi = solution[i];
	Vj = solution[j];
	
	switch (clueType) {
	  case Problem.SAME_COL:		
		if (Vi == Vj) {
		  result = true;
		  } 
		break;
	  case Problem.NOT_SAME_COL:
		if (Vi != Vj) {
		  result = true;
		}
		break;
	  case Problem.NEXT_TO:
		if (((Vi - Vj) ==  1) || 
			((Vi - Vj) == -1)) {
		  result = true;
		}
		break;	
	  case Problem.NOT_NEXT_TO:
		if (((Vi - Vj) !=  1) &&
			((Vi - Vj) != -1)) {
		  result = true;
		}
		break;
		case Problem.LEFT_OF:
		  if (Vi < Vj) {
			result = true;
		  }
		  break;
	  case Problem.RIGHT_OF:
		if (Vi > Vj) {
		  result = true;
		  }
		break;
	  case Problem.NOT_EQUAL:
		if (Vi != Vj) {
		  result = true;
		}
		break;
	  case Problem.NOT_NEXT_TO_SAME:
		if (((Vi - Vj) !=  1) &&
			((Vi - Vj) != -1) &&
			((Vi != Vj)))
		  {
			result = true;
		  }
		break;
	  case Problem.NEXT_RIGHT:
		if ((Vi - Vj) ==  1) {		  
		  result = true;
		  }
		break;
	  case Problem.NEXT_LEFT:
		if ((Vi - Vj) ==  -1) { 
		  result = true;
		  }
		break;
	  case Problem.GIVEN:
		result = true;
		break;	 
	}
	return result;
  }

  public  int getClueType() {
	double clueDouble = clueRand.nextDouble();
	double lowAllowedDouble = 0;
	double highAllowedDouble = ((Double)(allowedRelations[1].get(0))).doubleValue();
	for (int i = 0; i < allowedRelations[1].size(); i++) {
	  if (Debug.DEBUG_LEVEL >= Debug.DEBUG_VERBOSE) {
		System.out.println("i: " + String.valueOf(i) + " low: " +
						   String.valueOf(lowAllowedDouble) + " high: " +
						   String.valueOf(highAllowedDouble) + " rand: " +
						   String.valueOf(clueDouble));
	  }
	  if ((clueDouble >= lowAllowedDouble) && (clueDouble <= highAllowedDouble)) {
		int clueType = ((Integer)(allowedRelations[0].get(i))).intValue();
		return clueType;
	  }
	  if (i == (allowedRelations[1].size() - 1)) {
		return -2;
	  }
		lowAllowedDouble += ((Double)(allowedRelations[1].get(i))).doubleValue();
		highAllowedDouble += ((Double)(allowedRelations[1].get(i + 1))).doubleValue();
	}
	return -1;
  }

  public  Object getFromSet(Set set, int index) {
	int i = 0;
	Iterator itemIterator = set.iterator();
	Object item = null;

	while (itemIterator.hasNext() && i <= index) {
	  item = itemIterator.next();
	  i++;
	}
	if (Debug.DEBUG_LEVEL >= Debug.DEBUG_VERBOSE) {
	  System.out.println("Item " + item + " i: " + String.valueOf(i) + 
						 " index: " + String.valueOf(index));
	}

	return item;
  }

  public int findVar(int Row, int Col) {
	int varStart = (Row - 1) * problem.getMaxDomain() + 1;
	int varEnd = Row * problem.getMaxDomain();

	for (int i = varStart; i <= varEnd; i++) {
	  if (solution[i] == Col) {
		return i;
	  }
	}
	return 0;
  }

  public HashSet findAvail() {
	
	// Figure out number of times each column may be used
	int numColTimes = problem.getMaxDomain();
	int numColExtra = 0; // problem.getMaxDomain() % 2;
	
	// Figure out number of times each row may be used
	int numRowTimes = problem.getMaxDomain() - 1;
	int numRowExtra = problem.getMaxDomain() / 2;
	int maxExtraVarUse = 3;

	HashSet availPoints = new HashSet();
	int newPoint;

	for (int row = 1; row <= problem.getMaxDomain(); row++) {
	  int rowTimesUsed = ((Integer)(knownRows.get(new Integer(row)))).intValue();
	  if ((rowTimesUsed < numRowTimes) || ((rowTimesUsed == numRowTimes) && (rowExtraUsed < numRowExtra))) {
		for (int col = 1; col <= problem.getMaxDomain(); col++) {
		  int colTimesUsed = ((Integer)(knownCols.get(new Integer(col)))).intValue();
		  if ((colTimesUsed < numColTimes) || ((colTimesUsed == numColTimes) && (colExtraUsed < numColExtra))) {
			newPoint = findVar(row, col);
			if (newPoint == 0) {
			  if (Debug.DEBUG_LEVEL >= Debug.DEBUG_QUIET) {
				System.out.println("WARNING: Available variable set to zero.");
			  }
			}
			if (Debug.DEBUG_LEVEL >= Debug.DEBUG_HIGH) {
			  System.out.print(String.valueOf(newPoint) + " ");
			}
			if (knownVars.contains(new Integer(newPoint))) {
			  if (extraVarUse >= maxExtraVarUse) {
				continue;
			  }
			}
			availPoints.add(new Integer(newPoint));
		  } 
		}		
	  }
	}

	if (Debug.DEBUG_LEVEL >= Debug.DEBUG_HIGH) {
	  System.out.println("");
	}
	// 	done = false;
	if (availPoints.size() == 0) {
	  //	  done = true;
	  for (int i = 1; i <= problem.getMaxVars(); i++) {
		if (!(knownVars.contains(new Integer(i)))) {
		  availPoints.add(new Integer(i));
		  //		  done = false;
		}
	  }
	}
	//	if (done) {
	//	  return null;	
	//	} else {
	  return availPoints;
	  //	}
  }

  public HashSet findPossible() {
	HashSet possibleClues = new HashSet();
	int clueType = getClueType();
	int vx = 0;
	int vy = 0;
	HashSet availPoints = findAvail();
	//	if (availPoints == null) {
	//	  return null;
	//	}
	Iterator availIterator = availPoints.iterator();
	Iterator iterator2;
	boolean canConnect = false;
	while (availIterator.hasNext()) {
	  vx = ((Integer)(availIterator.next())).intValue();
	  if (Debug.DEBUG_LEVEL >= Debug.DEBUG_HIGH) {
		System.out.print("vx-avail: " + String.valueOf(vx) + " ");
	  }
	  iterator2 = ((HashSet)(availPoints.clone())).iterator();

	  while (iterator2.hasNext()) {
		vy = ((Integer)(iterator2.next())).intValue();
		if (vx == vy) {
		  if (Debug.DEBUG_LEVEL >= Debug.DEBUG_HIGH) {
			System.out.println("\nvx == vy == " + String.valueOf(vx));
		  }
		  continue;
		}
		if (checkAllowed(vx, vy, clueType)) {
		  if (Debug.DEBUG_LEVEL >= Debug.DEBUG_MAX) {
			System.out.println("Match on " + String.valueOf(vx) + " " +
							   String.valueOf(vy) + " " + 
							   String.valueOf(clueType));
		  }		  
		  possibleClues.add(new ClueObject(vx, clueType, vy));
		  canConnect = true;
		} else if (Debug.DEBUG_LEVEL >= Debug.DEBUG_HIGH) {
		  System.out.println("Fail on " + String.valueOf(vx) + " " +
							 String.valueOf(vy) + " " +
							 String.valueOf(clueType));
		}
	  }
// 	  while (!canConnect) {		
// 		vy = varRand.nextInt(problem.getMaxVars()) + 1;
// 		if (checkAllowed(vx, vy, clueType)) {
// 		  possibleClues.add(new ClueObject(vx, clueType, vy));
// 		  canConnect = true;
// 		}
// 	  }
	}
	if (possibleClues.size() == 0) {
	  if (Debug.DEBUG_LEVEL >= Debug.DEBUG_QUIET) {		
		System.out.println("Desired clue type not available; getting set of all clues.");
	  }
	  availIterator = availPoints.iterator();
	  while (availIterator.hasNext()) {
		vx = ((Integer)(availIterator.next())).intValue();
		if (Debug.DEBUG_LEVEL >= Debug.DEBUG_HIGH) {
		  System.out.print(" " + String.valueOf(vx) + " ");
		}
		iterator2 = ((HashSet)(availPoints.clone())).iterator();
		while (iterator2.hasNext()) {
		  vy = ((Integer)(iterator2.next())).intValue();
		  if (vx == vy) {
			continue;
		  }
		  Iterator allowedClueTypes = allowedRelations[0].iterator();
		  while (allowedClueTypes.hasNext()) {
			clueType = ((Integer)(allowedClueTypes.next())).intValue();
			if (checkAllowed(vx, vy, clueType)) {
			  possibleClues.add(new ClueObject(vx, clueType, vy));
			  canConnect = true;
			  if (Debug.DEBUG_LEVEL >= Debug.DEBUG_HIGH) {
				System.out.println(String.valueOf(vx) + " " + String.valueOf(vy) + " Type: " + String.valueOf(clueType));
			  }
			}
		  }
		}
		while (!canConnect) {		
		  //		  if (vx == vy) {
		  vy = varRand.nextInt(problem.getMaxVars()) + 1;
		  Iterator allowedClueTypes = allowedRelations[0].iterator();
		  while (allowedClueTypes.hasNext()) {
			clueType = ((Integer)(allowedClueTypes.next())).intValue();
			if (checkAllowed(vx, vy, clueType)) {
			  possibleClues.add(new ClueObject(vx, clueType, vy));
			  canConnect = true;
			  if (Debug.DEBUG_LEVEL >= Debug.DEBUG_HIGH) {
				System.out.println(String.valueOf(vx) + " " + String.valueOf(vy) + " Type: " + String.valueOf(clueType));
			  }			  
			}
		  }
		}
	  }
	  while (possibleClues.size() == 0) {
		vx = 0;
		vy = 0;
		while (vx == vy) {
		  vx = varRand.nextInt(problem.getMaxVars()) + 1;
		  vy = varRand.nextInt(problem.getMaxVars()) + 1;
		}
		Iterator allowedClueTypes = allowedRelations[0].iterator();
		while (allowedClueTypes.hasNext()) {
		  clueType = ((Integer)(allowedClueTypes.next())).intValue();
		  if (checkAllowed(vx, vy, clueType)) {
			possibleClues.add(new ClueObject(vx, clueType, vy));
			canConnect = true;
			if (Debug.DEBUG_LEVEL >= Debug.DEBUG_HIGH) {
			  System.out.println(String.valueOf(vx) + " " + String.valueOf(vy) + " Type: " + String.valueOf(clueType));
			}			  
		  }
		}		
	  }
	}
	return possibleClues;
 }

  public void addRow(int V) {
	int numRowTimes = problem.getMaxDomain() - 1;
	int row = problem.varRow(V) + 1;
	Integer rowCount = (Integer)(knownRows.get(new Integer(row)));
	if (rowCount != null) {
	  if ((rowCount.intValue() + 1) == numRowTimes) {
		rowExtraUsed++;
	  }
	  knownRows.put(new Integer(row), new Integer(rowCount.intValue() + 1));
	} else {
	  if (Debug.DEBUG_LEVEL >= Debug.DEBUG_QUIET) {
		System.out.println("First time for row " + String.valueOf(row));
	  }
	  knownRows.put(new Integer(row), new Integer(1));
	}
	if (Debug.DEBUG_LEVEL >= Debug.DEBUG_MEDIUM) {
	  System.out.println("Row " + String.valueOf(row) + " = " + 
						 knownCols.get(new Integer(row)));
	}
  }

  public void addCol(int V) {
	int numColTimes = problem.getMaxDomain();
	int numCols = problem.getMaxDomain();
	int col = solution[V];
	Integer colCount = (Integer)(knownCols.get(new Integer(col)));
	if (colCount != null) {
	  if ((colCount.intValue() + 1) >= numColTimes) {
		colExtraUsed++;
	  }
	  knownCols.put(new Integer(col), new Integer(colCount.intValue() + 1));
	} else {
	  if (Debug.DEBUG_LEVEL >= Debug.DEBUG_QUIET) {
		System.out.println("First time for col " + String.valueOf(col));
	  }
	  knownCols.put(new Integer(col), new Integer(1));
	}
	if (Debug.DEBUG_LEVEL >= Debug.DEBUG_MEDIUM) {
	  System.out.println("Col " + String.valueOf(col) + " = " + 
						 knownCols.get(new Integer(col)));
	}
  }

  public void addVar(int NewVar) {
	if (knownVars.contains(new Integer(NewVar))) {
	  extraVarUse++;
	} else {
	  knownVars.add(new Integer(NewVar));
	}
  }

  public void getClue() throws Exception {

	int vx;
	int vy;

	// Figure out possible clues
	HashSet possibleClues = findPossible();

	// If no possible clues we must be done, so exit
	//	if (possibleClues == null) {
	//	  done = true;
	//	  return;
	//	}
	if (possibleClues.size() == 0) {
	  System.err.println("Unable to find possible clue.");
	}

	int randChoice = varRand.nextInt(possibleClues.size());

	if (Debug.DEBUG_LEVEL >= Debug.DEBUG_MEDIUM) {
	  System.out.println("Choosing clue " + String.valueOf(randChoice) + 
						 " of " + String.valueOf(possibleClues.size()) + 
						 " clues.");

	}
	ClueObject newClue = (ClueObject)(getFromSet(possibleClues, randChoice));

	if (newClue.clueType == Problem.GIVEN) {
	  given[newClue.v2] = solution[newClue.v2];
	  problem.domain[newClue.v2] = new HashSet();
	  problem.domain[newClue.v2].add(new Integer(solution[newClue.v2]));
	  addCol(newClue.v2);
	  addVar(newClue.v1);
	} else {
	  problem.addConstraint(newClue.v1, newClue.v2, newClue.clueType);
	  addVar(newClue.v1);
	  addVar(newClue.v2);
	}

	if (Debug.DEBUG_LEVEL >= Debug.DEBUG_MEDIUM) {
	  System.out.println(String.valueOf(newClue.v1) + " " + String.valueOf(newClue.v2)  + " Type: " + String.valueOf(newClue.clueType));
	}

	switch (newClue.clueType) {
	  case Problem.SAME_COL:		
		problem.addConstraint(newClue.v2, newClue.v1, newClue.clueType);
		addRow(newClue.v1);
		addRow(newClue.v2);
		addCol(newClue.v1);
		addCol(newClue.v2);
		break;
	  case Problem.NOT_SAME_COL:
		problem.addConstraint(newClue.v2, newClue.v1, newClue.clueType);
		addCol(newClue.v1);
		break;
	  case Problem.NEXT_TO:
		problem.addConstraint(newClue.v2, newClue.v1, newClue.clueType);
		addCol(newClue.v2);
		addCol(newClue.v1);	   
		break;	
	  case Problem.NOT_NEXT_TO:
		problem.addConstraint(newClue.v2, newClue.v1, newClue.clueType);
		addCol(newClue.v1);
		addCol(newClue.v2);
		break;
	  case Problem.LEFT_OF:
		problem.addConstraint(newClue.v2, newClue.v1, Problem.RIGHT_OF);
		addCol(newClue.v1);
		addCol(newClue.v2);
		break;
	  case Problem.RIGHT_OF:
		problem.addConstraint(newClue.v2, newClue.v1, Problem.LEFT_OF);
		addCol(newClue.v1);
		addCol(newClue.v2);
		break;
	  case Problem.NOT_EQUAL:
		problem.addConstraint(newClue.v2, newClue.v1, newClue.clueType);
		addCol(newClue.v1);
		addCol(newClue.v2);
		break;
	  case Problem.NOT_NEXT_TO_SAME:
		problem.addConstraint(newClue.v2, newClue.v1, newClue.clueType);
		addCol(newClue.v1);
		addCol(newClue.v2);
		addRow(newClue.v1);
		addRow(newClue.v1);
		break;
	  case Problem.NEXT_RIGHT:
		problem.addConstraint(newClue.v2, newClue.v1, Problem.NEXT_LEFT);
		if (problem.varRow(newClue.v2) == problem.varRow(newClue.v1)) {
		  addRow(newClue.v1);
		} else {
		  addRow(newClue.v1);
		  addRow(newClue.v2);
		}
		addCol(newClue.v1);
		addCol(newClue.v2);
		break;
	  case Problem.NEXT_LEFT:
		problem.addConstraint(newClue.v2, newClue.v1, Problem.NEXT_RIGHT);
		if (problem.varRow(newClue.v2) == problem.varRow(newClue.v1)) {
		  addRow(newClue.v1);
		} else {
		  addRow(newClue.v1);
		  addRow(newClue.v2);
		}
		addCol(newClue.v1);
		addCol(newClue.v2);	
		break;
	}
  }

  public  Vector doPermute(TreeSet curDomain) {
	Vector permutations = new Vector();
	Vector integers;

	if (curDomain.size() == 1) {
	  integers = new Vector();
	  integers.addElement((Integer)(curDomain.first()));
	  permutations.addElement(integers);
	} else {	  
	  Iterator current = curDomain.iterator();	 
	  while (current.hasNext()) {
		TreeSet newDomain = (TreeSet)(curDomain.clone());
		Integer curInt = (Integer)(current.next());
		newDomain.remove(curInt);

		Vector endCombos = doPermute(newDomain);
		for (int i = 0; i < endCombos.size(); i++) {
		  integers = new Vector();
		  integers.addElement(curInt);
		  Vector endInts = (Vector)(endCombos.elementAt(i));

		  for (int j = 0; j < endInts.size(); j++) {
			integers.addElement((Integer)(endInts.elementAt(j)));
		  }
		  permutations.addElement(integers);
		}		
	  }	  
	}

	if (Debug.DEBUG_LEVEL >= Debug.DEBUG_MAX) {
	  for (int i = 0; i < permutations.size(); i++) {
		Vector ints = (Vector)(permutations.elementAt(i));
		System.out.print(String.valueOf(i) + ": ");
		for (int j = 0; j < ints.size(); j++) {
		  System.out.print(String.valueOf((Integer)(ints.elementAt(j)) + " "));
		}
		System.out.println("");
	  }
	}
	return permutations;
  }

  public  void setSolution(Object permVector, int varStart) {
	Vector permutes = (Vector)(permVector);
	for (int o = 0; o < permutes.size(); o++) {
	  solution[varStart + o] = 
		((Integer)(permutes.elementAt(o))).intValue();
	}
  }

  public  void generateSolution() {
	solution = new int[problem.getMaxVars() + 1];

	for (int i = 0; i <= problem.getMaxVars(); i++) {
	  solution[i] = Problem.NO_SOLUTION;
	}
	
	TreeSet topDomain = new TreeSet();
	Vector varPermutations;

	for (int i = 1; i <= problem.getMaxDomain(); i++) {
	  topDomain.add(new Integer(i));
	}

	varPermutations = doPermute(topDomain);

	for (int i = 0; i < problem.getMaxDomain(); i++) {
	  int x;
	  x = solutionRand.nextInt(varPermutations.size());
	  
	  setSolution(varPermutations.elementAt(x), i * problem.getMaxDomain() + 1 );
	}	
  }

  public void createUniqueConstraints() {
	HashSet rowSet;
	int v2;

	for (int i = 1; i <= problem.getMaxVars(); i++) {
	  rowSet = problem.getRowSet(problem.varRow(i));	  

	  Iterator rowIterator = rowSet.iterator();
	  while (rowIterator.hasNext()) {
		v2 = ((Integer)(rowIterator.next())).intValue();
		if (i != v2) {
		  problem.addConstraint(i, v2, Problem.NOT_EQUAL);
		}
	  }
	}
  }

  public String relationsString() {
	String out = "";

	if (type == Problem.SHERLOCK) {
	  out += "S\n";
	} else if (type == Problem.ZEBRA) {
	  out += "Z\n";
	}

	for (int i = 1; i <= problem.getMaxVars(); i++) {
	  if (given[i] != 0) {
		out += String.valueOf(i) + " is " + String.valueOf(given[i]) + "\n";
	  }
	}

	for (int i = 1; i <= problem.getMaxVars(); i++) {
	  for (int j = 1; j <= problem.getMaxVars(); j++) {
		if ((i != j) && (problem.constraints[i][j] != null)) {
		  out += problem.getConstraintString(i, j);
		}
	  }
	}
	return out;
  }

  class LabelResult {
	int i;
	boolean consistent;
	
	LabelResult(int I, boolean Consistent) {
	  i = I;
	  consistent = Consistent;
	}
  }

  public static HashSet currentDomain[];
	  
  public static final int UNKNOWN = 0;
  public static final int NO_SOLUTION = -1;
  public static final int SOLUTION = 1;
  public static final int MORE_THAN_ONE_SOLUTION = 2;

  public HashSet conflictSet[];
  public boolean cbf[];

  public int bcssp(int n) {
	boolean consistent = true;
	int status = UNKNOWN;
	int prevStatus = UNKNOWN;
	int i = 1;
	currentDomain = new HashSet[problem.getMaxVars() + 1];
	cbf = new boolean[problem.getMaxVars() + 1];
	for (int j = 0; j <= problem.getMaxVars(); j++) {
	  cbf[j] = false;
	}
	conflictSet = new HashSet[problem.getMaxVars() + 1];
	
	for (int j = 0; j <= problem.getMaxVars(); j++) {
	  conflictSet[j] = new HashSet();
	}	

	currentDomain[0] = null;

	for (int j = 1; j <= problem.getMaxVars(); j++) {
	  currentDomain[j] = (HashSet)(problem.domain[j].clone());
	}

	while (status == UNKNOWN) {
	  if ((Debug.DEBUG_LEVEL >= Debug.DEBUG_HIGH)) {
		System.out.println("Starting new check after " +
						   String.valueOf(problem.numChecks) + 
						   " checks and " + String.valueOf(problem.numNodes) +
						   " nodes.");
	  }

	  if (consistent) {
		LabelResult result = cbjLabel(i, consistent);
		i = result.i;
		consistent = result.consistent;
		//		problem.numNodes++;
	  } else {
		LabelResult result = cbjUnlabel(i, consistent);
		i = result.i;
		consistent = result.consistent;
	  }
	  if ((Debug.DEBUG_LEVEL >= Debug.DEBUG_HIGH)) {
		System.out.println("Checks: " + String.valueOf(problem.numChecks) + " " + ", Nodes: " + String.valueOf(problem.numNodes));
	  }

	  if (i > n) {
		if (prevStatus == UNKNOWN) {
		  prevStatus = SOLUTION;
		  for (int j = 0; j <= problem.getMaxVars(); j++) {
			cbf[j] = true;
		  }
   		  i = n;
		  consistent = false;
		} else {
		  if (prevStatus == SOLUTION) {
			status = MORE_THAN_ONE_SOLUTION;
		  } 
		}
	  } else if (i == 0) {
		if (prevStatus == SOLUTION) {
		  status = SOLUTION;
		} else {
		  status = NO_SOLUTION;
		}
	  }	  
	}
	return status;
  }

  public LabelResult cbjLabel(int i, boolean consistent) {
	LabelResult result;

	consistent = false;
	Iterator kIterator = currentDomain[i].iterator();
	HashSet toRemove = new HashSet();

	int k = 1;

	while ((!consistent) && kIterator.hasNext()) {	 
	  problem.numNodes++;
	  k = ((Integer)(kIterator.next())).intValue();
	  problem.solution[i] = k;
	  //	  cbf[i] = false;

	  consistent = true;

	  int h = 1;

	  while(consistent && (h <= i - 1)) {
		consistent = problem.check(i, h);
		h++;
	  }
	  if (!consistent) {
		conflictSet[i].add(new Integer(h - 1));
		toRemove.add(new Integer(problem.solution[i]));
	  }	  
	}
	if (consistent) {
	  result = new LabelResult(i + 1, consistent);
	  //	  cbf[i] = false;
	} else {
	  result = new LabelResult(i, consistent);
	}

	Iterator lIterator = toRemove.iterator();
	while (lIterator.hasNext()) {	  
	  currentDomain[i].remove(lIterator.next());		
	}
	return result;
  }
  public LabelResult cbjUnlabel(int i, boolean consistent) {
	LabelResult result;
	int h = 1;

	if (cbf[i]) {
	  h = i - 1;
	} else {
	  h = maxSet(conflictSet[i]);
	
	  if (h > 0) {
		conflictSet[h].addAll(conflictSet[i]);
		conflictSet[h].remove(new Integer(h));
	  }
	}
	
	for (int j = h + 1; j <= i; j++) {
	  conflictSet[j] = new HashSet();
	  currentDomain[j] = (HashSet)(problem.domain[j].clone());
	  cbf[j] = false;
	}
	  
	if (h > 0) {
	  currentDomain[h].remove(new Integer(problem.solution[h]));
	  consistent = !(currentDomain[h].isEmpty());
	}
  
	result = new LabelResult(h, consistent);
	cbf[i] = false;
	return result;
  }

  public int maxSet(HashSet S) {
	Iterator i = S.iterator();
	int j;
	int maxJ = 0;

	while (i.hasNext()) {
	  j = ((Integer)i.next()).intValue();
	  if (j > maxJ) {
		maxJ = j;
	  }
	}
	return maxJ;
  }
}


