// Copyright 2002-2003 University of Guelph, All Rights Reserved
import java.util.*;

class XoverChromosone implements Comparable, Cloneable {

  public static final double MUTATE_PROBABILITY = 0.01;

  int chromosone[];
  int maxDomain;
  Problem problem;
  double fitness;
//   boolean badRows[];

  XoverChromosone(Problem ProblemObject) {
	problem = ProblemObject;
	chromosone = new int[problem.getMaxVars() + 1];
	maxDomain = problem.getMaxDomain();
	fitness = 0.0;
  }
  
  XoverChromosone(int Chromosone[], Problem ProblemObject) {
	chromosone = (int[])(Chromosone.clone());
	chromosone[0] = 0;
	fitness = 0;
	problem = ProblemObject;
	maxDomain = problem.getMaxDomain();
	//	badRows = new boolean[problem.getMaxDomain()];	
	calcFitness();
  }

  public void reproduce(XoverChromosone Parent1, 
								   XoverChromosone Parent2,
						XoverChromosone Sibling2,
								   Random XoverRandom) {	
	fitness = 0;
	problem = Parent1.problem;
	maxDomain = problem.getMaxDomain();
	chromosone = new int[problem.getMaxVars() + 1];
	chromosone[0] = 0;
	//	int xoverI = XoverRandom.nextInt(problem.getMaxDomain());
	//	int xoverPoint = maxDomain * xoverI + 1;
	int xoverPoint = XoverRandom.nextInt(problem.getMaxVars()) + 1;
	for (int i = 1; i < xoverPoint; i++) {
	  chromosone[i] = Parent1.chromosone[i];
	  Sibling2.chromosone[i] = Parent2.chromosone[i];
	}
	for (int i = xoverPoint; i <= problem.getMaxVars(); i++) {
	  chromosone[i] = Parent2.chromosone[i];
	  Sibling2.chromosone[i] = Parent1.chromosone[i];
	}	
  }

  public void mutate(Random mutateRandom, int numAlleles) {
// 	HashSet rowSet;
// 	Vector rows = MutateOnly.rows;
//     Vector newRow;
//  	Integer newCol;

// 	double mutateProbability = (double)(problem.getMaxDomain()) / (double)(problem.getMaxVars()) * 4;

// 	int numRowsToMutate = (int)Math.ceil((mutateProbability * mutateRandom.nextDouble() * problem.getMaxDomain()));
// 	HashSet rowsToMutate = new HashSet();
// // 	for (int i = 0; i < maxDomain; i++) {
// 	  if (badRows[i]) {
// 		rowsToMutate.add(new Integer(i));
// 	  }
// 	}
//     int rowToMutate;
// 	for (int i = 0; i < numRowsToMutate; i++) {
// 	  if (rowsToMutate.size() > 0) {
// 		rowToMutate = ((Integer)(Problem.getFromSet(rowsToMutate, mutateRandom.nextInt(rowsToMutate.size())))).intValue();
// 		rowsToMutate.remove(new Integer(rowToMutate));
// 	  } else {
// 		rowToMutate = mutateRandom.nextInt(maxDomain);
// 	  }
// 	  newRow = (Vector)(rows.elementAt(mutateRandom.nextInt(rows.size())));
//   	  for (int j = 0; j < newRow.size(); j++) {
// 		newCol = (Integer)(newRow.elementAt(j));
// 		chromosone[rowToMutate * maxDomain + j + 1] = newCol.intValue();
//  	  }	
// 	}	


	for (int i = 1; i <= problem.getMaxVars(); i++) {
	  if (mutateRandom.nextDouble() < MUTATE_PROBABILITY) {
		chromosone[i] = mutateRandom.nextInt(maxDomain) + 1;
	  }
	}
	
	calcFitness();
  }
  
  protected void calcFitness() {
	int i , j;
 	int numConflicts;
 	fitness = 0.0;	

   	for (i = 1; i <= problem.getMaxVars(); i++) {
	  if (!(problem.domain[i].contains(new Integer(chromosone[i])))) {
		fitness += 200.0;		
	  } else {
 		for (j = 1; j <= problem.getMaxVars(); j++) {
		  if (!(problem.checkConsistent(chromosone[i], chromosone[j], 
								i, j))) {
			Iterator iterator = problem.constraints[i][j].iterator();
			int distance = chromosone[i] - chromosone[j];
			
			while (iterator.hasNext()) {
			  switch(((Integer)(iterator.next())).intValue()) {
				case Problem.SAME_COL:
				  fitness += Math.abs(distance);
				  break;
				case Problem.NOT_SAME_COL:
				  fitness += 1.0;
				  break;
				case Problem.NEXT_TO:
				  if (distance == 0) {
					fitness += 1.0;
				  } else {
					fitness += Math.abs(distance);
				  }
				  break;	
				case Problem.NOT_NEXT_TO:
				  fitness += 1.0;
				  break;
				case Problem.LEFT_OF:
				  if (distance == 0) {
					fitness += 1.0;
				  } else {
					fitness += Math.abs(distance);
				  }
				  break;
				case Problem.RIGHT_OF:
				  if (distance == 0) {
					fitness += 1.0;
				  } else {
					fitness += Math.abs(distance);
				  }
				  break;
				case Problem.NOT_EQUAL:
				  fitness += 25.0;
				  break;
				case Problem.NOT_NEXT_TO_SAME:
				  if (distance == 0) {
					fitness += 1.0;
				  } else {
					fitness += Math.abs(distance);
				  }
				  break;
				case Problem.NEXT_RIGHT:
				  if (distance == 0) {
					fitness += 1.0;
				  } else {
					fitness += Math.abs(distance);
				  }
				  break;
				case Problem.NEXT_LEFT:
				  if (distance == 0) {
					fitness += 1.0;
				  } else {
					fitness += Math.abs(distance);
				  }
				  break;
			  }
			}
		  }			  
		}
	  } 
	}
	
// 	problem.solution = ((int[])(chromosone.clone()));
// 	numConflicts = problem.checkConflicts();
//  	fitness += (double)(numConflicts * 1.0);

// 	if (numConflicts == 0) {
// 	  fitness = 0.0;
// 	}

	if (Debug.DEBUG_LEVEL >= Debug.DEBUG_VERBOSE) {
	  System.out.println("                                                                   " + String.valueOf(fitness));
	}
  }

  public Object clone() { // throws CloneNotSupportedException {
	SimpleChromosone newObject = new SimpleChromosone(chromosone, problem);
	newObject.fitness = fitness;
	return newObject;
  }
  
  public int[] getSolution() {
	return chromosone;
  }
  
  public double getFitness()
  {
	return fitness;
  }
  
  public boolean isSolution() {
	problem.solution = (int[])(chromosone.clone());
	return (problem.checkSolution());
  }
  
  public int compareTo(Object o) throws ClassCastException {
	int result = 0;
	boolean sameChromosone = true;
	
	if (o.getClass() != this.getClass()) {
	  throw new ClassCastException("Trying to compare SimpleChromosone to " +
								   o.getClass().getName());
	}
	
	XoverChromosone rhs = (XoverChromosone)(o);
	
	if (getFitness() < rhs.getFitness()) {
	  result = -1;
	} else if (getFitness() == rhs.getFitness()) {
	  for (int i = 0; i < chromosone.length; i++) {
		if (chromosone[i] < rhs.chromosone[i]) {
		  sameChromosone = false;
		  result = -1;
		  break;
		} else if (chromosone[i] > rhs.chromosone[i]) {
		  sameChromosone = false;
		  result = 1;
		  break;
		}
	  }

	  if (sameChromosone) {
		result = 0;
	  }
	} else if (getFitness() > rhs.getFitness()) {
	  result = 1;
	}  
	return result;
  }

  public boolean equals(Object rhs) {
	return (compareTo(rhs) == 0);
  }

  public String toString() {
	String out = "";
	
	for (int i = 0; i < chromosone.length; i++) {
	  out += String.valueOf(chromosone[i]);
	}

	out += " : " + String.valueOf(getFitness());

	return out;
  }

  public String getSolutionString() {
	problem.solution = (int[])(chromosone.clone());
	return problem.getSolutionString();
  }
}
